from django.utils.deprecation import MiddlewareMixin

from atris.models import HistoryLogging


class LoggingRequestMiddleware(MiddlewareMixin):
    def process_request(self, request):
        HistoryLogging.thread.request = request



# -*- coding: utf-8 -*-
# Generated by Django 1.11.8 on 2018-01-12 11:41
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('atris', '0007_related_field_history'),
    ]

    operations = [
        migrations.AlterField(
            model_name='archivedhistoricalrecord',
            name='object_id',
            field=models.TextField(db_index=True),
        ),
        migrations.AlterField(
            model_name='historicalrecord',
            name='object_id',
            field=models.TextField(db_index=True),
        ),
    ]



from django.apps import AppConfig


class AtrisConfig(AppConfig):
    name = "atris"
    label = "atris"
    verbose_name = "Atris Model History"

    def ready(self):
        from atris.models import registered_models

        for sender in registered_models:
            # We have access to all the fields of a Django model only after all
            # models have been loaded.
            history_logger = sender._meta.history_logging
            history_logger.set_additional_data_properties(sender)
            history_logger.set_excluded_fields_names(sender)
            history_logger.set_interested_related_fields(sender)
            history_logger.register_signal_handlers(sender)



from django.contrib import admin
from django.contrib.contenttypes.models import ContentType
from django.db import connections, models
from django.urls import reverse
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _

from atris.models import ArchivedHistoricalRecord, HistoricalRecord, history_logging


class ContentTypeListFilter(admin.SimpleListFilter):
    # Human-readable title which will be displayed in the
    # right admin sidebar just above the filter options.
    title = _("Content type")

    # Parameter for the filter that will be used in the URL query.
    parameter_name = "content_type"

    def lookups(self, request, model_admin):
        """
        Returns a list of tuples. The first element in each
        tuple is the coded value for the option that will
        appear in the URL query. The second element is the
        human-readable name for the option that will appear
        in the right sidebar.
        """
        followed_content_types = [
            ContentType.objects.get_for_model(model)
            for model in history_logging.registered_models.keys()
        ]

        filter_results = set()
        for content_type in followed_content_types:
            filter_results.add((content_type.id, _(content_type.model)))

        return filter_results

    def queryset(self, request, queryset):
        """
        Returns the filtered queryset based on the value
        provided in the query string and retrievable via
        `self.value()`.
        """
        if self.value() is not None:
            return queryset.filter(content_type__id=self.value())


class ApproxCountPgQuerySet(models.query.QuerySet):
    """approximate unconstrained count(*) with reltuples from pg_class"""

    def count(self):
        if hasattr(connections[self.db].client.connection, "pg_version"):
            query = self.query
            no_filtration_used = (
                not query.where
                and query.high_mark is None
                and query.low_mark == 0
                and not query.select
                and not query.group_by
                and not query.having
                and not query.distinct
            )
            if no_filtration_used:
                parts = [p.strip('"') for p in self.model._meta.db_table.split(".")]
                if 1 <= len(parts) <= 2:
                    cursor = connections[self.db].cursor()
                    if len(parts) == 1:
                        cursor.execute(
                            "SELECT reltuples::bigint "
                            "FROM pg_class "
                            "WHERE relname = %s",
                            parts,
                        )
                    else:
                        cursor.execute(
                            "SELECT reltuples::bigint "
                            "FROM pg_class c "
                            "JOIN pg_namespace n on (c.relnamespace = n.oid) "
                            "WHERE n.nspname = %s AND c.relname = %s",
                            parts,
                        )
                    return cursor.fetchall()[0][0]
        return self.query.get_count(using=self.db)


class GenericHistoryAdmin(admin.ModelAdmin):
    list_display = (
        "object_id",
        "content_type",
        "history_date",
        "history_type",
        "history_user",
    )

    fields = [
        "object_id",
        "content_type",
        "history_date",
        "history_type",
        "history_user",
        "difference_to_previous",
        "fields_that_differ",
        "history_snapshot",
        "more_info",
        "related_field_history_admin",
    ]

    readonly_fields = fields

    search_fields = ("object_id",)

    list_filter = (ContentTypeListFilter, "history_type")

    show_full_result_count = False

    def get_queryset(self, request):
        # Capturing the request object in order to build the absolute URI in
        # `related_field_history_admin`
        self._request = request
        qs = super().get_queryset(request)
        return qs

    def history_snapshot(self, obj):
        return self._dict_to_table(obj.data)

    def more_info(self, obj):
        return self._dict_to_table(obj.additional_data)

    def difference_to_previous(self, obj):
        return obj.get_diff_to_prev_string()

    def fields_that_differ(self, obj):
        if obj.history_diff:
            return ", ".join(obj.history_diff)
        return None

    def _dict_to_table(self, dictionary):
        table = (
            '<table style="border: 1px solid #eee;">'
            + "".join(
                [
                    '<tr><td style="border: 1px solid #eee;">{}</td> <td>{}'
                    "</td></tr>".format(
                        key,
                        val,
                    )
                    for (key, val) in dictionary.items()
                ]
            )
            + "</table>"
        )

        return mark_safe(table)

    def related_field_history_admin(self, obj):
        if obj.related_field_history:
            related_url = reverse(
                "admin:atris_historicalrecord_change",
                args=[obj.related_field_history.pk],
            )
            absolute_uri = self._request.build_absolute_uri(related_url)
            related_object_model = obj.related_field_history.content_type.model
            html = '<a href="{}">{}</a>'.format(
                absolute_uri,
                obj.additional_data[related_object_model],
            )
            return mark_safe(html)
        else:
            return "--"

    related_field_history_admin.short_description = "Related Field History"

    def has_add_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False


class HistoricalRecordAdmin(GenericHistoryAdmin):
    model = HistoricalRecord
    show_full_result_count = False


class ArchivedHistoricalRecordAdmin(GenericHistoryAdmin):
    model = ArchivedHistoricalRecord
    show_full_result_count = False


admin.site.register(HistoricalRecord, HistoricalRecordAdmin)
admin.site.register(ArchivedHistoricalRecord, ArchivedHistoricalRecordAdmin)



from django.conf import settings
from django.contrib.contenttypes.models import ContentType

from .abstract_historical_record import AbstractHistoricalRecord


class HistoricalRecord(AbstractHistoricalRecord):
    pass


def get_history_model():
    try:
        app_label, model = settings.ATRIS_HISTORY_MODEL.lower().split(".")
    except AttributeError:
        return HistoricalRecord
    else:
        content_type = ContentType.objects.get_by_natural_key(app_label, model)
        return content_type.model_class()



import logging

from django.core.management import BaseCommand

from atris.models import ArchivedHistoricalRecord, get_history_model


logger = logging.getLogger("old_history_deleting")
HistoricalRecord = get_history_model()


class Command(BaseCommand):
    help = """
        Deletes historical records older than the specified days or months.
        You must supply either the days or the weeks param.
    """

    PARAM_ERROR = "You must supply either the days or the weeks param"

    def add_arguments(self, parser):
        parser.add_argument(
            "--days",
            dest="days",
            type=int,
            default=None,
            help=(
                "Any historical record older than "
                "the number of days specified gets deleted."
            ),
        )
        parser.add_argument(
            "--weeks",
            dest="weeks",
            type=int,
            default=None,
            help=(
                "Any historical record older than "
                "the number of months specified gets deleted."
            ),
        )
        parser.add_argument(
            "--from-archive",
            dest="from_archive",
            default=False,
            action="store_true",
            help=(
                'Delete occurs on the "archived historical records" table '
                'instead of the default "historical records" table.'
            ),
        )

    def handle(self, *args, **options):
        days = options.get("days")
        weeks = options.get("weeks")
        if not (days or weeks):
            self.stderr.write(f"{self.PARAM_ERROR}\n")
            return

        model = (
            ArchivedHistoricalRecord
            if options.get("from_archive")
            else HistoricalRecord
        )
        deleted_entries = model.objects.older_than(days, weeks).delete()
        self.stdout.write(f"{deleted_entries[0]} {model.__name__} deleted.\n")



# -*- coding: utf-8 -*-
# Generated by Django 1.10.5 on 2017-03-02 09:27
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('atris', '0006_auto_20160622_0720'),
    ]

    operations = [
        migrations.AddField(
            model_name='archivedhistoricalrecord',
            name='related_field_history',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='referenced_objects_history', to='atris.ArchivedHistoricalRecord'),
        ),
        migrations.AddField(
            model_name='historicalrecord',
            name='related_field_history',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='referenced_objects_history', to='atris.HistoricalRecord'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10 on 2019-09-06 05:09
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('atris', '0008_object_id_text'),
    ]

    operations = [
        migrations.AlterField(
            model_name='archivedhistoricalrecord',
            name='object_id',
            field=models.TextField(),
        ),
        migrations.AlterField(
            model_name='historicalrecord',
            name='object_id',
            field=models.TextField(),
        ),
        migrations.AlterIndexTogether(
            name='archivedhistoricalrecord',
            index_together=set([('object_id', 'history_date')]),
        ),
        migrations.AlterIndexTogether(
            name='historicalrecord',
            index_together=set([('object_id', 'history_date')]),
        ),
    ]



from math import ceil

from django.core.management import BaseCommand
from django.db import DatabaseError
from django.db.transaction import atomic

from atris.models import get_history_model, registered_models
from atris.models.helpers import get_instance_field_data


HistoricalRecord = get_history_model()


class Command(BaseCommand):
    EXISTING_HISTORY_FOUND = "Existing history found, skipping model"

    SELECT_BATCH_SIZE = 1000
    CREATE_BATCH_SIZE = 1000

    def add_arguments(self, parser):
        parser.add_argument(
            "--select-batch-size",
            dest="select_batch_size",
            type=int,
            default=self.SELECT_BATCH_SIZE,
            help=(
                "The number of target objects that "
                "will be retrieved from the DB at one time."
            ),
        )
        parser.add_argument(
            "--create-batch-size",
            dest="create_batch_size",
            type=int,
            default=self.CREATE_BATCH_SIZE,
            help=(
                "The number of history objects that will be created "
                "in one batch. Should be at most SELECT_BATCH_SIZE."
            ),
        )

    def handle(self, *args, **options):
        for model, model_specific_info in registered_models.items():
            if HistoricalRecord.objects.by_model(model).exists():
                self.stderr.write(f"{self.EXISTING_HISTORY_FOUND} {model}\n")
                continue
            self.stdout.write(f"Initializing history for {model}\n")
            additional_data_param_name = model_specific_info[
                "additional_data_param_name"
            ]
            additional_data_field = getattr(
                model,
                additional_data_param_name,
                {},
            )
            excluded_fields_param_name = model_specific_info[
                "excluded_fields_param_name"
            ]
            excluded_fields = getattr(model, excluded_fields_param_name, [])
            create_history_for_model = ModelHistoryCreator(
                model,
                additional_data_field,
                excluded_fields,
                options["select_batch_size"],
                options["create_batch_size"],
                self.stdout,
            )
            try:
                with atomic():
                    create_history_for_model()
            except DatabaseError as e:
                self.stderr.write(
                    f"Error creating history for {model}: {e}",
                )


class ModelHistoryCreator:
    def __init__(
        self,
        model,
        additional_data_field,
        excluded_fields,
        select_batch_size,
        create_batch_size,
        output,
    ):
        self.model = model
        self.additional_data_field = additional_data_field
        self.excluded_fields = excluded_fields
        self.select_batch_size = select_batch_size
        self.create_batch_size = create_batch_size
        self.output = output

    def __call__(self):
        objects = self.model.objects
        number_of_batches = ceil(objects.count() / self.select_batch_size)
        self.output.write(
            "Processing data in {} batches of {} target objects.\n".format(
                number_of_batches,
                self.select_batch_size,
            ),
        )
        for multiplicity in range(number_of_batches):
            start = self.select_batch_size * multiplicity
            end = start + self.select_batch_size
            self.create_history_for_objects(objects.all()[start:end])
            self.output.write(
                "Finished batch #{} of {}.\n".format(
                    multiplicity + 1,
                    number_of_batches,
                ),
            )

    def create_history_for_objects(self, objects):
        historical_instances = []
        for instance in objects:
            historical_record = self.create_history_for_object(instance)
            historical_instances.append(historical_record)
        HistoricalRecord.objects.bulk_create(
            historical_instances,
            batch_size=self.create_batch_size,
        )

    def create_history_for_object(self, obj):
        data = get_instance_field_data(obj)
        additional_data = {
            key: str(value) for key, value in self.additional_data_field.items()
        }
        historical_record = HistoricalRecord(
            history_user=None,
            history_type=HistoricalRecord.CREATE,
            content_object=obj,
            data=data,
            additional_data=additional_data,
        )
        return historical_record



# Generated by Django 4.2.6 on 2023-10-16 11:46

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('atris', '0009_auto_20190906_0509'),
    ]

    operations = [
        migrations.AlterField(
            model_name='archivedhistoricalrecord',
            name='additional_data',
            field=models.JSONField(null=True),
        ),
        migrations.AlterField(
            model_name='archivedhistoricalrecord',
            name='data',
            field=models.JSONField(),
        ),
        migrations.AlterField(
            model_name='historicalrecord',
            name='additional_data',
            field=models.JSONField(null=True),
        ),
        migrations.AlterField(
            model_name='historicalrecord',
            name='data',
            field=models.JSONField(),
        ),
    ]



import ast
import json
import re

from typing import Dict, List, Optional, Type, Union

from django.contrib.contenttypes.fields import GenericForeignKey, GenericRelation
from django.core.exceptions import ObjectDoesNotExist
from django.db import router
from django.db.models import Model


def get_diff_fields(
    model: Union[Model, Type[Model]],
    data: Dict,
    previous_data: Dict,
    excluded_fields_names: List[str],
) -> Optional[List[str]]:
    """

    Args:
        model: the Django model or an instance of that model
        data: current instance data as dictionary
        previous_data: previous instance data as dictionary
        excluded_fields_names: excluded fields names of the specified model

    Returns: the fields of `data` for which the values differ in

    """
    if not previous_data:
        return None
    diff_fields = []
    for f, v in data.items():
        if f in excluded_fields_names:
            continue
        model_field = model._meta.get_field(f)
        try:
            previous_value = previous_data[f]
        except KeyError:
            if v != get_default_value(model_field):
                diff_fields.append(model_field.name)
        else:
            is_field_different = is_different(
                old=previous_value,
                new=v,
                field=model_field,
            )
            if is_field_different:
                diff_fields.append(model_field.name)
    return diff_fields


def get_default_value(field):
    try:
        default_value = field.get_default()
        # this is an ugly fix needed for django 3.2 compatibility - I don't like it
        # either but have no better solution
        if isinstance(field, GenericRelation) and default_value is None:
            return ''
        return str(default_value) if default_value is not None else None
    except AttributeError as e:
        error_message = str(e)
        if "ManyTo" in error_message:
            # ManyToManyRel or ManyToOneRel
            return ""
        return None


def get_field_internal_type(field):
    try:
        return field.get_internal_type().strip()
    except AttributeError as e:
        if type(field) == GenericForeignKey:
            return "GenericForeignKey"
        else:
            raise e


def is_different(old, new, field):
    # serializing/deserializing a json object can change key order
    id_list = re.compile(r"^(\d+,\s)+\d+$")
    new_list = re.match(id_list, new or "")
    prev_list = re.match(id_list, old or "")
    is_list = new_list and prev_list
    is_relation_to_many = (field.one_to_many or field.many_to_many) and is_list

    field_internal_type = get_field_internal_type(field=field)
    # django jsonfield allows both python dicts or raw json
    if field_internal_type == "JSONField":
        try:
            # for valid python dict
            old = ast.literal_eval(old) if old else old
        except ValueError:
            # for string of json
            old = json.loads(old) if old else old
        try:
            new = ast.literal_eval(new) if new else new
        except ValueError:
            new = json.loads(new) if new else new
    elif is_relation_to_many or field_internal_type == "ArrayField":
        old = set(old.split(", ")) if old else old
        new = set(new.split(", ")) if new else new
    return old != new


def get_instance_field_data(instance):
    """
    Returns a dictionary with the attribute values of instance, serialized as
    strings.
    """
    data = {}
    instance_meta = instance._meta
    for field in instance_meta.get_fields():
        name = field.name
        if name in instance_meta.history_logging.excluded_fields_names:
            continue
        attname = get_attribute_name_from_field(field)
        try:
            value = getattr(instance, attname)
        except ObjectDoesNotExist:
            value = None
        if field.many_to_many or field.one_to_many:
            ids = from_writable_db(value).values_list("pk", flat=True)
            data[name] = ", ".join([str(e) for e in ids.order_by("pk")])
        elif field.one_to_one and not field.concrete:
            data[name] = str(value.pk) if value is not None else None
        else:
            data[name] = str(value) if value is not None else None
    return data


def get_attribute_name_from_field(field, flat_fk=True):
    accessor_for_simple_fields = "attname" if flat_fk else "name"
    if hasattr(field, "fk_field"):  # generic foreign key
        attname = field.fk_field
    elif hasattr(field, "get_accessor_name"):  # many-to-* relation field
        attname = field.get_accessor_name()
    elif hasattr(field, accessor_for_simple_fields):
        # regular field or foreign key
        attname = getattr(field, accessor_for_simple_fields)
    else:
        raise TypeError(f"Can't determine accessor for field {field}")
    return attname


def from_writable_db(manager):
    """
    When using a DB router in which the reads are done through a different
    connection than the writes, the data may differ, resulting in incorrect
    history logging. Use the writable DB whenever it is essential to get the
    latest data.
    """
    writable_db = router.db_for_write(manager.model)
    return manager.using(writable_db)
